import numpy as np
import numpy.linalg as la
import scipy.linalg as sla
import scipy.special
import matplotlib.pyplot as plt
import scipy.sparse.linalg as spla
from operators import *
##########################################################################################
### RCIP functions:
def Pmat_corner(t16,w16):
    # Obtain Pbc matrix for corner
    t32 = np.concatenate((t16-1,t16+1))/2
    w32 = np.concatenate((w16,w16))/2
    A, _, _ = vander_op(t16)
    B, _, _ = vander_op(t32)
    B = B[:,0:np.shape(A)[0]]
    BP = B @ la.inv(A)
    BPW = w32.reshape(-1,1) * BP / w16.reshape(1,-1)
    Pbc1 = sla.block_diag(np.eye(np.size(t16)),BP,np.rot90(BP,2),np.eye(np.size(t16)))
    PbcW1 = sla.block_diag(np.eye(np.size(t16)),BPW,np.rot90(BPW,2),np.eye(np.size(t16)))
    return Pbc1, PbcW1
def Pmat_intersect(t16,w16):
    # Obtain Pbc matrix for intersection
    t32 = np.concatenate((t16-1,t16+1))/2
    w32 = np.concatenate((w16,w16))/2
    A, _, _ = vander_op(t16)
    B, _, _ = vander_op(t32)
    B = B[:,0:np.shape(A)[0]]
    BP = B @ la.inv(A)
    BPW = w32.reshape(-1,1) * BP / w16.reshape(1,-1)
    Pbc1 = sla.block_diag(np.eye(np.size(t16)),BP,np.rot90(BP,2),np.eye(np.size(t16)),np.eye(np.size(t16)),BP)
    PbcW1 = sla.block_diag(np.eye(np.size(t16)),BPW,np.rot90(BPW,2),np.eye(np.size(t16)),np.eye(np.size(t16)),BPW)
    return Pbc1, PbcW1
def R_crn(crn_num,nsub,nref,quad_info,geom_info,zg,nzg,wg):
    bp, tp, side, A = geom_info.intersect, geom_info.tip, geom_info.side, geom_info.area
    n = quad_info.shape[1]
    Nstar = 6*n
    star_idx = np.concatenate((np.arange(n,5*n),np.arange(Nstar+n,Nstar+5*n),np.arange(2*Nstar+n,2*Nstar+5*n),np.arange(3*Nstar+n,3*Nstar+5*n)))
    rnodes_g = (scipy.special.legendre(2*n).weights)[:,0]
    rw_g = (scipy.special.legendre(2*n).weights)[:,2]
    crn_idx, locm1, locm2 = 0, 0, 0
    if crn_num == 1:
        crn_idx, locm1, locm2 = quad_info.idx_crn_1, 1j, 1
    elif crn_num == 2:
        crn_idx, locm1, locm2 = quad_info.idx_crn_2, -1, 1j
    elif crn_num == 3:
        crn_idx, locm1, locm2 = quad_info.idx_crn_3, -1j, -1
    elif crn_num == 4:
        crn_idx, locm1, locm2 = quad_info.idx_crn_4, 1, -1j
    else:
        raise ValueError("R_crn(crn_num,nsub,quad_info,geom_info,zg,nzg,wg): crn_num not implemented!!")
    pl = quad_info.panel_lengths[crn_idx]
    rnodes_c = quad_info.ref_nodes[crn_idx,:]
    rw_c = quad_info.ref_weights[crn_idx,:]
    rweights_c1, rweights_c2 = (quad_info.curve_weights_plain.reshape(quad_info.shape))[crn_idx-1,:], (quad_info.curve_weights_plain.reshape(quad_info.shape))[crn_idx,:]
    idcs = np.arange(n*(crn_idx-2), n*(crn_idx+2)) + n*nref
    zc = quad_info.panel_nodes[crn_idx] # Corner Point
    zglob = np.delete(zg,idcs) # Global Points of the rest of the panels
    nzglob = np.delete(nzg,idcs) # Global Normals of the rest of the panels
    wglob = np.delete(wg,idcs) # Global Weights of the rest of the panels
    K = np.zeros((6,np.shape(quad_info.diffmat)[1],np.shape(quad_info.diffmat)[2])) + quad_info.diffmat[crn_idx,:,:]
    Pbc_c, Pbcw_c = Pmat_corner(rnodes_c,rw_c)
    Pbc_c, Pbcw_c = sla.block_diag(Pbc_c,Pbc_c,Pbc_c,Pbc_c), sla.block_diag(Pbcw_c,Pbcw_c,Pbcw_c,Pbcw_c)
    for i in range(nsub):
        d = 2**(nsub - i - 1)
        ## Global Leftovers:
        zleftover11, zleftover21 = zc + locm1 * 2/d * pl, zc + locm1 * 2 * pl
        zleftover12, zleftover22 = zc + locm2 * 2/d * pl, zc + locm2 * 2 * pl
        int_length1, int_length2 = abs(zleftover11 - zleftover21), abs(zleftover12 - zleftover22)
        zp = np.concatenate(((zleftover11 + zleftover21)/2 + (zleftover11 - zleftover21) * rnodes_g/2,(zleftover12 + zleftover22)/2 + (zleftover22 - zleftover12) * rnodes_g/2 ))
        nzp = np.concatenate((rnodes_g*0 + 1j*locm1, rnodes_g*0 -1j*locm2))
        wp1 = np.concatenate((-locm1 * rw_g * int_length1/2, locm2  * rw_g * int_length2/2))/rho(zp,bp,tp,side)
        # Global Quantities:
        zg = np.concatenate((zglob, zp))
        nzg = np.concatenate((nzglob, nzp))
        wg = np.concatenate((wglob.reshape(-1), wp1.reshape(-1)))
        ## Local Quantities:
        locpar = np.concatenate((rnodes_c+1,rnodes_c+3,2*rnodes_c+6))/(4*d)
        zloc = np.concatenate((zc + locm1 * np.flipud(locpar * pl),zc + locm2 * locpar * pl))
        nzloc = np.concatenate((locpar*0 - locm2,locpar*0 - locm1))
        wloc = np.concatenate((np.flipud(np.concatenate((rweights_c1,rweights_c1,2*rweights_c1))),
                               np.concatenate((rweights_c2,rweights_c2,2*rweights_c2))))/(4*d)
        rholoc = rho(zloc,bp,tp,side)
        wl1, wl2 = (wloc / rholoc).reshape(1,-1), (wloc * rholoc).reshape(1,-1)
        ## Local Operators:
        M31c, M32c = M3_rho(zloc,nzloc,wl2)
        M1c_inv = M1_rho_inv_loc(zloc,nzloc,wl1,zg,nzg,wg.reshape(1,-1),K,n)
        Qzc = Q(zloc,wl2)
        iP0_1c,iP0_2c = iP_0(zloc,wl2,side,A)
        hzc = h(zloc,side)
        M1c_r, _ = Op_to_real(M1c_inv)
        M31c_r, _ = Op_to_real(M31c)
        _, M32c_r = Op_to_real(M32c)
        hzbarQc_r, _ = Op_to_real(np.conj(hzc*zloc).reshape(-1,1)*Qzc)
        hiP01c_r, _ = Op_to_real(hzc.reshape(-1,1)*iP0_1c)
        _, hiP02c_r = Op_to_real(hzc.reshape(-1,1)*iP0_2c)
        Opc_l = -(M31c_r + M32c_r + hzbarQc_r + hiP01c_r + hiP02c_r)
        # Composed Operator:
        Odim = np.shape(Opc_l)[0]
        MK = np.zeros((2*Odim,2*Odim))
        MK[0:Odim,Odim:2*Odim] = Opc_l
        MK[Odim:2*Odim,0:Odim] = -M1c_r
        Op = np.eye(2*Odim) + MK
        # Recursion for R
        if i == 0:
            R = la.inv(Op[np.ix_(star_idx,star_idx)])
        Op[np.ix_(star_idx,star_idx)] = la.inv(R)
        R = Pbcw_c.T @ la.inv(Op) @ Pbc_c
    Rdim = int(np.shape(R)[0]/2)
    R1 = R[0:Rdim,0:Rdim]
    R2 = R[Rdim:2*Rdim,0:Rdim]
    R3 = R[0:Rdim,Rdim:2*Rdim]
    R4 = R[Rdim:2*Rdim,Rdim:2*Rdim]
    R13= la.inv(R1) @ R3;
    R4213= R4 - R2 @ R13;
    idcs2 = np.concatenate((idcs, quad_info.shape[0]*quad_info.shape[1] + idcs))
    return idcs2, R1, R2, R13, R4213
def R_int(nsub,nref,quad_info,geom_info,zg,nzg,wg):
    bp, tp, side, A = geom_info.intersect, geom_info.tip, geom_info.side, geom_info.area
    n = quad_info.shape[1]
    Nstar = 9*n
    star_idx = np.concatenate((np.arange(n,5*n),np.arange(7*n,9*n),
                               np.arange(Nstar+n,Nstar+5*n),np.arange(Nstar+7*n,Nstar+9*n),
                               np.arange(2*Nstar+n,2*Nstar+5*n),np.arange(2*Nstar+7*n,2*Nstar+9*n),
                               np.arange(3*Nstar+n,3*Nstar+5*n),np.arange(3*Nstar+7*n,3*Nstar+9*n)))
    rnodes_g = (scipy.special.legendre(2*n).weights)[:,0]
    rw_g = (scipy.special.legendre(2*n).weights)[:,2]
    int_idx = quad_info.idx_intersect
    pl = quad_info.panel_lengths[int_idx]
    rnodes_i = quad_info.ref_nodes[int_idx,:]
    rw_i = quad_info.ref_weights[int_idx,:]
    rweights_i1, rweights_i2, rweights_i3 = (quad_info.curve_weights_plain.reshape(quad_info.shape))[int_idx-1,:], (quad_info.curve_weights_plain.reshape(quad_info.shape))[int_idx,:], (quad_info.curve_weights_plain.reshape(quad_info.shape))[-1,:]
    idcs = np.concatenate((np.arange(n*(int_idx-2), n*(int_idx+2)), np.arange(n*(quad_info.shape[0]-2), (quad_info.shape[0]*n)))) + n*nref
    zi = quad_info.panel_nodes[int_idx] # Intersection Point
    zglob = np.delete(zg,idcs) # Global Points of the rest of the panels
    nzglob = np.delete(nzg,idcs) # Global Normals of the rest of the panels
    wglob = np.delete(wg,idcs) # Global Weights of the rest of the panels
    K = np.zeros((9,np.shape(quad_info.diffmat)[1],np.shape(quad_info.diffmat)[2])) + quad_info.diffmat[int_idx,:,:]
    Pbc_i, Pbcw_i = Pmat_intersect(rnodes_i,rw_i)
    Pbc_i, Pbcw_i = sla.block_diag(Pbc_i,Pbc_i,Pbc_i,Pbc_i), sla.block_diag(Pbcw_i,Pbcw_i,Pbcw_i,Pbcw_i)
    for i in range(nsub):
        d = 2**(nsub - i - 1)
        ## Global Leftovers:
        zleftover11, zleftover21 = zi + 2/d * pl, zi + 2 * pl
        zleftover12, zleftover22 = zi - 1j * 2/d * pl, zi - 1j * 2 * pl
        zleftover13, zleftover23 = zi + 1j * 2/d * pl, zi + 1j * 2 * pl
        int_length1, int_length2, int_length3 = abs(zleftover11 - zleftover21), abs(zleftover12 - zleftover22), abs(zleftover13 - zleftover23)
        zp = np.concatenate(((zleftover11 + zleftover21)/2 + (zleftover11 - zleftover21) * rnodes_g/2,(zleftover12 + zleftover22)/2 + (zleftover22 - zleftover12) * rnodes_g/2,(zleftover13 + zleftover23)/2 + (zleftover13 - zleftover23) * rnodes_g/2 ))
        nzp = np.concatenate((rnodes_g*0 + 1j, rnodes_g*0 -1, rnodes_g*0 -1))
        wp = np.concatenate((-1 * rw_g * int_length1/2, -1j  * rw_g * int_length2/2, -1j  * rw_g * int_length3/2))/rho(zp,bp,tp,side)
        # Global Quantities:
        zg = np.concatenate((zglob, zp))
        nzg = np.concatenate((nzglob, nzp))
        wg = np.concatenate((wglob.reshape(-1), wp.reshape(-1)))
        ## Local Quantities:
        locpar = np.concatenate((rnodes_i+1,rnodes_i+3,2*rnodes_i+6))/(4*d)
        zloc = np.concatenate((zi + np.flipud(locpar * pl),zi - 1j * locpar * pl, zi + 1j * np.flipud(locpar * pl)))
        nzloc = np.concatenate((0*locpar + 1j, 0*locpar - 1, 0*locpar - 1))
        wloc = np.concatenate((np.flipud(np.concatenate((rweights_i1,rweights_i1,2*rweights_i1))),
                               np.concatenate((rweights_i2,rweights_i2,2*rweights_i2)),
                               np.flipud(np.concatenate((rweights_i3,rweights_i3,2*rweights_i3)))))/(4*d)
        rholoc = rho(zloc,bp,tp,side)
        wl1, wl2 = (wloc / rholoc).reshape(1,-1), (wloc * rholoc).reshape(1,-1)
        ## Local Operators:
        M31i, M32i = M3_rho(zloc,nzloc,wl2)
        M1i_inv = M1_rho_inv_loc(zloc,nzloc,wl1,zg,nzg,wg.reshape(1,-1),K,n)
        Qzi = Q(zloc,wl2)
        iP0_1i,iP0_2i = iP_0(zloc,wl2,side,A)
        hzi = h(zloc,side)
        M1i_r, _ = Op_to_real(M1i_inv)
        M31i_r, _ = Op_to_real(M31i)
        _, M32i_r = Op_to_real(M32i)
        hzbarQi_r, _ = Op_to_real(np.conj(hzi*zloc).reshape(-1,1)*Qzi)
        hiP01i_r, _ = Op_to_real(hzi.reshape(-1,1)*iP0_1i)
        _, hiP02i_r = Op_to_real(hzi.reshape(-1,1)*iP0_2i)
        Opi_l = -(M31i_r + M32i_r + hzbarQi_r + hiP01i_r + hiP02i_r)
        # Composed Operator:
        Odim = np.shape(Opi_l)[0]
        MK = np.zeros((2*Odim,2*Odim))
        MK[0:Odim,Odim:2*Odim] = Opi_l
        MK[Odim:2*Odim,0:Odim] = -M1i_r
        Op = np.eye(2*Odim) + MK
        # Recursion for R:
        if i == 0:
            R = la.inv(Op[np.ix_(star_idx,star_idx)])
        Op[np.ix_(star_idx,star_idx)] = la.inv(R)
        R = Pbcw_i.T @ la.inv(Op) @ Pbc_i
    Rdim = int(np.shape(R)[0]/2)
    R1 = R[0:Rdim,0:Rdim]
    R2 = R[Rdim:2*Rdim,0:Rdim]
    R3 = R[0:Rdim,Rdim:2*Rdim]
    R4 = R[Rdim:2*Rdim,Rdim:2*Rdim]
    R13= la.inv(R1) @ R3;
    R4213= R4 - R2 @ R13;
    idcs2 = np.concatenate((idcs, quad_info.shape[0]*quad_info.shape[1] + idcs))
    return idcs2, R1, R2, R13, R4213
def zero_entries(F, idcs1, idcs2, idcs3, idcs4, idcsi):
    H = np.copy(F)
    H[np.ix_(idcs1,idcs1)] = 0
    H[np.ix_(idcs2,idcs2)] = 0
    H[np.ix_(idcs3,idcs3)] = 0
    H[np.ix_(idcs4,idcs4)] = 0
    H[np.ix_(idcsi,idcsi)] = 0
    return H