import numpy as np
import matplotlib.pyplot as plt
import scipy.sparse.linalg as spla
from operators import *
from quad import *
from RCIP_operators import *
##########################################################################################
### Important!! RCIP Method working correctly, however, obtaining the fine solution
# density for it is not implemented. To appreciate the real benefits of the method
# for our purposes we need to obtain the fine solution back from the coarse RCIP solution
# This is done recursively in a similar fashion than the RCIP method itself...
##########################################################################################
### Definition of Geometry and Mesh Quantities Square domain with edge crack
side = 16 # Side length
npanels = 9*4  # Number of coarse panels of equal size used (has to be of the form 9k)
npts = 16 # Number of points per panel (Defines the quadrature rule)
nref = 0 # Number of Refinement towards the tip
nsub = 0 # Number of RCIP Recursions
##########################################################################################
class gmres_counter(object):
    # Can print iteration number and residual for each iteration using 
    # scipy.sparse.linalg.gmres when disp=True
    # Borrowed from stackoverflow answer:
    # https://stackoverflow.com/questions/33512081/getting-the-number-of-iterations-of-scipys-gmres-iterative-method
    def __init__(self, disp=True):
        self._disp = disp
        self.niter = 0
    def __call__(self, rk=None):
        self.niter += 1
        if self._disp:
            print('iter %3i\trk = %s' % (self.niter, str(rk)))
##########################################################################################
### Definition of the used variables for Coarse Boundary Problem:
geom_info = GeometricInfo(side)
A = geom_info.area
bp = geom_info.intersect
tp = geom_info.tip
quad_info = QuadratureInfo(npanels,npts,geom_info)
z = quad_info.curve_nodes
nz = quad_info.normals
dz = quad_info.curve_deriv
w = quad_info.curve_weights_plain.reshape(1,-1)
w1 = quad_info.curve_weights_rho_inv.reshape(1,-1)
w2 = quad_info.curve_weights_rho.reshape(1,-1)
tr = neumann_bcs(z,nz,side)
K = quad_info.diffmat
r = rho(z,bp,tp,side)
### Obtaining the fine mesh points:
if nsub > 0:
    w = w.reshape(-1)
    w1 = w1.reshape(-1)
    w2 = w2.reshape(-1)
    zf_c1,nzf_c1,wf_c1,rhof_c1, Kf_c1 = refine_crn(quad_info.idx_crn_1,nsub,quad_info,geom_info)
    zf_c2,nzf_c2,wf_c2,rhof_c2, Kf_c2 = refine_crn(quad_info.idx_crn_2,nsub,quad_info,geom_info)
    zf_c3,nzf_c3,wf_c3,rhof_c3, Kf_c3 = refine_crn(quad_info.idx_crn_3,nsub,quad_info,geom_info)
    zf_c4,nzf_c4,wf_c4,rhof_c4, Kf_c4 = refine_crn(quad_info.idx_crn_4,nsub,quad_info,geom_info)
    zf_i0,nzf_i0,wf_i0,rhof_i0, Kf_i0,zf_i1,nzf_i1,wf_i1,rhof_i1, Kf_i1 = refine_int(nsub,quad_info,geom_info)
    
    inter1 = np.arange(0,npts*(quad_info.idx_intersect-1))
    inter2 = np.arange(npts*(quad_info.idx_intersect+1),npts*(quad_info.idx_crn_1-1))
    inter3 = np.arange(npts*(quad_info.idx_crn_1+1),npts*(quad_info.idx_crn_2-1))
    inter4 = np.arange(npts*(quad_info.idx_crn_2+1),npts*(quad_info.idx_crn_3-1))
    inter5 = np.arange(npts*(quad_info.idx_crn_3+1),npts*(quad_info.idx_crn_4-1))
    inter6 = np.arange(npts*(quad_info.idx_crn_4+1),npts*(npanels-1))

    zf = np.concatenate((z[inter1],zf_i0,z[inter2],zf_c1,z[inter3],zf_c2,z[inter4],zf_c3,z[inter5],zf_c4,z[inter6],zf_i1))
    nzf = np.concatenate((nz[inter1],nzf_i0,nz[inter2],nzf_c1,nz[inter3],nzf_c2,nz[inter4],nzf_c3,nz[inter5],nzf_c4,nz[inter6],nzf_i1))
    wf = (np.concatenate((w[inter1],wf_i0,w[inter2],wf_c1,w[inter3],wf_c2,w[inter4],wf_c3,w[inter5],wf_c4,w[inter6],wf_i1))).reshape(1,-1)
    w1f = (np.concatenate((w1[inter1],wf_i0/rhof_i0,w1[inter2],wf_c1/rhof_c1,w1[inter3],wf_c2/rhof_c2,w1[inter4],wf_c3/rhof_c3,w1[inter5],wf_c4/rhof_c4,w1[inter6],wf_i1/rhof_i1))).reshape(1,-1)
    w2f = (np.concatenate((w2[inter1],wf_i0*rhof_i0,w2[inter2],wf_c1*rhof_c1,w2[inter3],wf_c2*rhof_c2,w2[inter4],wf_c3*rhof_c3,w2[inter5],wf_c4*rhof_c4,w2[inter6],wf_i1*rhof_i1))).reshape(1,-1)
    Kf = np.concatenate((K[(inter1/npts).astype(int),:,:],Kf_i0,K[(inter2/npts).astype(int),:,:],Kf_c1,K[(inter3/npts).astype(int),:,:],Kf_c2,K[(inter4/npts).astype(int),:,:],Kf_c3,K[(inter5/npts).astype(int),:,:],Kf_c4,K[(inter6/npts).astype(int),:,:],Kf_i1),axis=0)
    w = w.reshape(1,-1)
    w1 = w1.reshape(1,-1)
    w2 = w2.reshape(1,-1)
### Adaptive Refinement of Tip:
if nref > 0:
    w = w.reshape(-1)
    w1 = w1.reshape(-1)
    w2 = w2.reshape(-1)
    zf_t,nzf_t,wf_t,rhof_t, Kf_t = refine_tip(nref,quad_info,geom_info)
    z = np.concatenate((zf_t,z[npts:]))
    nz = np.concatenate((nzf_t,nz[npts:]))
    w = (np.concatenate((wf_t,w[npts:]))).reshape(1,-1)
    w1 = (np.concatenate((wf_t/rhof_t,w1[npts:]))).reshape(1,-1)
    w2 = (np.concatenate((wf_t*rhof_t,w2[npts:]))).reshape(1,-1)
    tr = neumann_bcs(z,nz,side)
    K = np.concatenate((Kf_t,K[1:,:,:]),axis=0)
    r = rho(z,bp,tp,side)
    if nsub > 0:
        wf = wf.reshape(-1)
        w1f = w1f.reshape(-1)
        w2f = w2f.reshape(-1)
        zf = np.concatenate((zf_t,zf[npts:]))
        nzf = np.concatenate((nzf_t,nzf[npts:]))
        wf = (np.concatenate((wf_t,wf[npts:]))).reshape(1,-1)
        w1f = (np.concatenate((wf_t/rhof_t,w1f[npts:]))).reshape(1,-1)
        w2f = (np.concatenate((wf_t*rhof_t,w2f[npts:]))).reshape(1,-1)
        trf = neumann_bcs(zf,nzf,side)
        Kf = np.concatenate((Kf_t,Kf[1:,:,:]),axis=0)
        rf = rho(zf,bp,tp,side)
### Definition of the Operators for Coarse Boundary Problem:
M31, M32 = M3_rho(z,nz,w2)
M1_inv = M1_rho_inv(z,nz,w1,K,npanels,npts)
gz = g(z,nz,tr,w,bp,tp)
Qz = Q(z,w2)
iP0_1,iP0_2 = iP_0(z,w2,side,A)
hz = h(z,side)
M1_r, _ = Op_to_real(M1_inv)
M31_r, _ = Op_to_real(M31)
_, M32_r = Op_to_real(M32)
hzbarQ_r, _ = Op_to_real(np.conj(hz*z).reshape(-1,1)*Qz)
hiP01_r, _ = Op_to_real(hz.reshape(-1,1)*iP0_1)
_, hiP02_r = Op_to_real(hz.reshape(-1,1)*iP0_2)
gz_r = vec_to_real(gz)
Op_l = -(M31_r + M32_r + hzbarQ_r + hiP01_r + hiP02_r)
##########################################################################################
### RCIP Method
R1 = np.eye(np.shape(Op_l)[0])
R2 = np.zeros((np.shape(Op_l)[0],np.shape(Op_l)[1]))
R13 = np.zeros((np.shape(Op_l)[0],np.shape(Op_l)[1]))
R4213 = np.eye(np.shape(Op_l)[0])
if nsub > 0:
    ## Assemble the contributions of each corner and junction:
    #Corner 1:
    idcs_c1, R1_c, R2_c, R13_c, R4213_c = R_crn(1,nsub,nref,quad_info,geom_info,z,nz,w1)
    R1[np.ix_(idcs_c1,idcs_c1)] = R1_c
    R2[np.ix_(idcs_c1,idcs_c1)] = R2_c
    R13[np.ix_(idcs_c1,idcs_c1)] = R13_c
    R4213[np.ix_(idcs_c1,idcs_c1)] = R4213_c
    #Corner 2:
    idcs_c2, R1_c, R2_c, R13_c, R4213_c = R_crn(2,nsub,nref,quad_info,geom_info,z,nz,w1)
    R1[np.ix_(idcs_c2,idcs_c2)] = R1_c
    R2[np.ix_(idcs_c2,idcs_c2)] = R2_c
    R13[np.ix_(idcs_c2,idcs_c2)] = R13_c
    R4213[np.ix_(idcs_c2,idcs_c2)] = R4213_c
    #Corner 3:
    idcs_c3, R1_c, R2_c, R13_c, R4213_c = R_crn(3,nsub,nref,quad_info,geom_info,z,nz,w1)
    R1[np.ix_(idcs_c3,idcs_c3)] = R1_c
    R2[np.ix_(idcs_c3,idcs_c3)] = R2_c
    R13[np.ix_(idcs_c3,idcs_c3)] = R13_c
    R4213[np.ix_(idcs_c3,idcs_c3)] = R4213_c
    #Corner 4:
    idcs_c4, R1_c, R2_c, R13_c, R4213_c = R_crn(4,nsub,nref,quad_info,geom_info,z,nz,w1)
    R1[np.ix_(idcs_c4,idcs_c4)] = R1_c
    R2[np.ix_(idcs_c4,idcs_c4)] = R2_c
    R13[np.ix_(idcs_c4,idcs_c4)] = R13_c
    R4213[np.ix_(idcs_c4,idcs_c4)] = R4213_c
    #Intersection:
    idcs_i, R1_i, R2_i, R13_i, R4213_i= R_int(nsub,nref,quad_info,geom_info,z,nz,w1)
    R1[np.ix_(idcs_i,idcs_i)] = R1_i
    R2[np.ix_(idcs_i,idcs_i)] = R2_i
    R13[np.ix_(idcs_i,idcs_i)] = R13_i
    R4213[np.ix_(idcs_i,idcs_i)] = R4213_i
    #
Op_l_circ = zero_entries(Op_l, idcs_c1, idcs_c2, idcs_c3, idcs_c4, idcs_i)
M1_r_circ = zero_entries(M1_r, idcs_c1, idcs_c2, idcs_c3, idcs_c4, idcs_i)
O = np.eye(np.shape(Op_l)[0]) + Op_l_circ @ R4213 @ M1_r_circ @ R1 + Op_l_circ @ R2 - R13 @ M1_r_circ @ R1
Op =  np.eye(np.shape(Op_l)[0]) + Op_l @ M1_r 
### Solve the Boundary Problem:
counter = gmres_counter()
Omtil, _ = spla.gmres(O,gz_r,tol=1e-13, callback=counter,restart=100)
### Recovering the coarse (real) density:
S = la.inv(Op) @ O
Omega = S @ Omtil
### Recover coarse (complex) density:
Om =     vec_to_complex(Omega)
sigma = (r * ((M1_inv) * Om.reshape(1,-1)).sum(axis=1))
### Obtain the stresses at domain:
# Mesh of target points:
grid_res = 200
grid_lim = 10
z_target = (np.linspace(-grid_lim,grid_lim,grid_res).reshape(-1,1) + 1j * np.linspace(-grid_lim,grid_lim,grid_res).reshape(1,-1)).reshape(-1)
X = (z_target.real).reshape(grid_res,grid_res)
Y = (z_target.imag).reshape(grid_res,grid_res)
# Stress related operators:
PH = PHI(z,z_target,sigma,w)
PS = PSI(z,nz,tr,z_target,sigma,w)
PHPR = PHIPR(z,z_target,sigma,w)
# Stresses:
sxx = np.real(2*PH - z_target*np.conj(PHPR) - np.conj(PS))
syy = np.real(2*PH + z_target*np.conj(PHPR) + np.conj(PS))
sxy = -np.imag(z_target*np.conj(PHPR) + np.conj(PS))
von = vonMises(sxx,syy,sxy)
### Plot of results:
levels = np.linspace(0,3.5,20)
CS = plt.contourf(X,Y,von.reshape(grid_res,grid_res), levels,origin='lower', extend='max')
CS.cmap.set_over('yellow')
CS1 = plt.contour(X,Y,von.reshape(grid_res,grid_res), levels,colors=('k',),linewidths=(0.5,),origin='lower')
cbar = plt.colorbar(CS)
cbar.ax.set_ylabel('vonMises Stress')
cbar.add_lines(CS1)